/*
C   SUBROUTINE NAME     - CUBGCV
C   USAGE               - CALL CUBGCV (X,F,DF,N,Y,C,IC,VAR,JOB,SE,WK,IER)
C   ARGUMENTS    X      - VECTOR OF LENGTH N CONTAINING THE
C                           ABSCISSAE OF THE N DATA POINTS
C                           (X(I),F(I)) I=1..N. (INPUT) X
C                           MUST BE ORDERED SO THAT
C                           X(I) .LT. X(I+1).
C                F      - VECTOR OF LENGTH N CONTAINING THE
C                           ORDINATES (OR FUNCTION VALUES)
C                           OF THE N DATA POINTS (INPUT).
C                DF     - VECTOR OF LENGTH N. (INPUT/OUTPUT)
C                           DF(I) IS THE RELATIVE STANDARD DEVIATION
C                           OF THE ERROR ASSOCIATED WITH DATA POINT I.
C                           EACH DF(I) MUST BE POSITIVE.  THE VALUES IN
C                           DF ARE SCALED BY THE SUBROUTINE SO THAT
C                           THEIR MEAN SQUARE VALUE IS 1, AND UNSCALED
C                           AGAIN ON NORMAL EXIT.
C                           THE MEAN SQUARE VALUE OF THE DF(I) IS RETURNED
C                           IN WK(7) ON NORMAL EXIT.
C                           IF THE ABSOLUTE STANDARD DEVIATIONS ARE KNOWN,
C                           THESE SHOULD BE PROVIDED IN DF AND THE ERROR
C                           VARIANCE PARAMETER VAR (SEE BELOW) SHOULD THEN
C                           BE SET TO 1.
C                           IF THE RELATIVE STANDARD DEVIATIONS ARE UNKNOWN,
C                           SET EACH DF(I)=1.
C                N      - NUMBER OF DATA POINTS (INPUT).
C                           N MUST BE .GE. 3.
C                Y,C    - SPLINE COEFFICIENTS. (OUTPUT) Y
C                           IS A VECTOR OF LENGTH N. C IS
C                           AN N-1 BY 3 MATRIX. THE VALUE
C                           OF THE SPLINE APPROXIMATION AT T IS
C                           S(T)=((C(I,3)*D+C(I,2))*D+C(I,1))*D+Y(I)
C                           WHERE X(I).LE.T.LT.X(I+1) AND
C                           D = T-X(I).
C                IC     - ROW DIMENSION OF MATRIX C EXACTLY
C                           AS SPECIFIED IN THE DIMENSION
C                           STATEMENT IN THE CALLING PROGRAM. (INPUT)
C                VAR    - ERROR VARIANCE. (INPUT/OUTPUT)
C                           IF VAR IS NEGATIVE (I.E. UNKNOWN) THEN
C                           THE SMOOTHING PARAMETER IS DETERMINED
C                           BY MINIMIZING THE GENERALIZED CROSS VALIDATION
C                           AND AN ESTIMATE OF THE ERROR VARIANCE IS
C                           RETURNED IN VAR.
C                           IF VAR IS NON-NEGATIVE (I.E. KNOWN) THEN THE
C                           SMOOTHING PARAMETER IS DETERMINED TO MINIMIZE
C                           AN ESTIMATE, WHICH DEPENDS ON VAR, OF THE TRUE
C                           MEAN SQUARE ERROR, AND VAR IS UNCHANGED.
C                           IN PARTICULAR, IF VAR IS ZERO, THEN AN
C                           INTERPOLATING NATURAL CUBIC SPLINE IS CALCULATED.
C                           VAR SHOULD BE SET TO 1 IF ABSOLUTE STANDARD
C                           DEVIATIONS HAVE BEEN PROVIDED IN DF (SEE ABOVE).
C                JOB    - JOB SELECTION PARAMETER. (INPUT)
C                         JOB = 0 SHOULD BE SELECTED IF POINT STANDARD ERROR
C                           ESTIMATES ARE NOT REQUIRED IN SE.
C                         JOB = 1 SHOULD BE SELECTED IF POINT STANDARD ERROR
C                           ESTIMATES ARE REQUIRED IN SE.
C                SE     - VECTOR OF LENGTH N CONTAINING BAYESIAN STANDARD
C                           ERROR ESTIMATES OF THE FITTED SPLINE VALUES IN Y.
C                           SE IS NOT REFERENCED IF JOB=0. (OUTPUT)
C                WK     - WORK VECTOR OF LENGTH 7*(N + 2). ON NORMAL EXIT THE
C                           FIRST 7 VALUES OF WK ARE ASSIGNED AS FOLLOWS:-
C
C                           WK(1) = SMOOTHING PARAMETER (= RHO/(RHO + 1))
C                           WK(2) = ESTIMATE OF THE NUMBER OF DEGREES OF
C                                   FREEDOM OF THE RESIDUAL SUM OF SQUARES
C                           WK(3) = GENERALIZED CROSS VALIDATION
C                           WK(4) = MEAN SQUARE RESIDUAL
C                           WK(5) = ESTIMATE OF THE TRUE MEAN SQUARE ERROR
C                                   AT THE DATA POINTS
C                           WK(6) = ESTIMATE OF THE ERROR VARIANCE
C                           WK(7) = MEAN SQUARE VALUE OF THE DF(I)
C
C                           IF WK(1)=0 (RHO=0) AN INTERPOLATING NATURAL CUBIC
C                           SPLINE HAS BEEN CALCULATED.
C                           IF WK(1)=1 (RHO=INFINITE) A LEAST SQUARES
C                           REGRESSION LINE HAS BEEN CALCULATED.
C                           WK(2) IS AN ESTIMATE OF THE NUMBER OF DEGREES OF
C                           FREEDOM OF THE RESIDUAL WHICH REDUCES TO THE
C                           USUAL VALUE OF N-2 WHEN A LEAST SQUARES REGRESSION
C                           LINE IS CALCULATED.
C                           WK(3),WK(4),WK(5) ARE CALCULATED WITH THE DF(I)
C                           SCALED TO HAVE MEAN SQUARE VALUE 1.  THE
C                           UNSCALED VALUES OF WK(3),WK(4),WK(5) MAY BE
C                           CALCULATED BY DIVIDING BY WK(7).
C                           WK(6) COINCIDES WITH THE OUTPUT VALUE OF VAR IF
C                           VAR IS NEGATIVE ON INPUT.  IT IS CALCULATED WITH
C                           THE UNSCALED VALUES OF THE DF(I) TO FACILITATE
C                           COMPARISONS WITH A PRIORI VARIANCE ESTIMATES.
C
C                IER    - ERROR PARAMETER. (OUTPUT)
C                         TERMINAL ERROR
C                           IER = 129, IC IS LESS THAN N-1.
C                           IER = 130, N IS LESS THAN 3.
C                           IER = 131, INPUT ABSCISSAE ARE NOT
C                             ORDERED SO THAT X(I).LT.X(I+1).
C                           IER = 132, DF(I) IS NOT POSITIVE FOR SOME I.
C                           IER = 133, JOB IS NOT 0 OR 1.
C
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "mex.h"

void CubVal(double *X,double *Y,double *C[3],double *nX,double *nY,int N,int nN),
     CubGCV(double *X,double *F,double *DF,int N,double *Y,double *C[3],
           double VAR,int Job,double *SE),
     SpInt(double *X,double *AvH,double *Y,double *dY,double *AvdY,int N,
          double *A,double *C[3],double *R[3],double *T[2]),
     SpFit(double *X,double AvH,double *dY,int N,double Rho,double *P,double *Q,
           double *Fun,double Var,double *Stat,double *A,double *C[3],double *R[3],
           double *T[2],double *U,double *V),
     SpErr(double *X,double AvH,double *dY,int N,double *R[3],double P,
           double Var,double *SE),
     SpCof(double *X,double AvH,double *Y,double *dY,int N,double P,double Q,
           double *A,double *C[3],double *U,double *V);

void mexFunction(int nl, mxArray *pl[], int nr, const mxArray *pr[]) {
  double *X, *F, *DF, *Y, *C[3], VAR, *SE, *nX, *nY;
  int     i, N, nN, Job;

  if ((nl < 1) || (4 < nl)) mexErrMsgTxt("One to four output arguments are allowed.");
  if (nr == 3) {
    N = mxGetM(pr[0]); nN = mxGetM(pr[2]); F = mxGetPr(pr[1]);
    if (mxGetN(pr[0]) != 1) mexErrMsgTxt("x have to be a column vector.");
    if (mxGetN(pr[2]) != 1) mexErrMsgTxt("new-x have to be a column vector.");
    if (N != mxGetM(pr[1])) mexErrMsgTxt("y have to have same number of rows as x.");
    if (mxGetN(pr[1]) == 2) {
      DF = F+N; VAR = 1.0; Job = 1;
    } else if (mxGetN(pr[1]) == 1) {
      DF = mxCalloc(N,sizeof(double)); VAR = 0.0; Job = 0;
      for (i = 0; i < N; i++) DF[i] = 1.0;
    } else mexErrMsgTxt("y have to be a column vector or Nx2 matrix.");
    X = mxGetPr(pr[0]); nX = mxGetPr(pr[2]);
  } else mexErrMsgTxt("Three input arguments are allowed.");
  Y = mxCalloc(N,sizeof(double)); nY = mxCalloc(nN,sizeof(double));
  SE = mxCalloc(N,sizeof(double)); C[0] = mxCalloc(3*(N-1),sizeof(double));
  C[1] = C[0]+N-1; C[2] = C[1]+N-1;

  CubGCV(X,F,DF,N,Y,C,VAR,Job,SE); CubVal(X,Y,C,nX,nY,N,nN);

  if (nl) {
    pl[0] = mxCreateDoubleMatrix(nN,1,mxREAL); mxSetPr(pl[0],nY);
    if (1 < nl) {
      pl[1] = mxCreateDoubleMatrix(N,1,mxREAL); mxSetPr(pl[1],Y);
      if (2 < nl) {
        pl[2] = mxCreateDoubleMatrix(N-1,3,mxREAL); mxSetPr(pl[2],C[0]);
        if (3 < nl) {
          pl[3] = mxCreateDoubleMatrix(N,1,mxREAL); mxSetPr(pl[3],SE);
        } else { mxFree(SE); }
      } else { mxFree(SE); mxFree(C[0]); }
    } else { mxFree(SE); mxFree(C[0]); mxFree(Y); }
  } else { mxFree(SE); mxFree(C[0]); mxFree(Y); mxFree(nY); }
}

void CubVal(double *X,double *Y,double *C[3],double *nX,double *nY,int N,int nN) {
  int    i = 0, j = 0;
  double d, mx, my;

  mx = X[0]; my = Y[0];
  while ((i < nN) && (nX[i] < mx)) nY[i++] = my;
  mx = X[N-1]; my = Y[N-1]; N -= 2;
  while ((i < nN) && (nX[i] < mx)) {
    while ((X[j+1] < nX[i]) && (j < N)) j++;
    d = nX[i]-X[j]; nY[i++] = Y[j]+d*(C[0][j]+d*(C[1][j]+d*C[2][j]));
  }
  while (i < nN) nY[i++] = my;
}

void CubGCV(double *X,double *F,double *DF,int N,double *Y,double *C[3],
           double VAR,int Job,double *SE) {
  int    i, More, GoTo100 = 0;
  double Delta, Err, GF1, GF2, GF3, GF4, R1, R2, R3, R4, Tau = 1.618033989,
         Ratio = 2.0, AVH, AVDF, AVAR, STAT[6], P, Q, ERR, *WK[7];

  if ((Job < 0) || (Job > 1)) mexErrMsgTxt("Job is not 0 or 1.");
  WK[0] = mxCalloc(7*(N+2),sizeof(double));
  WK[1] = &WK[0][1*(N+2)]; WK[2] = &WK[0][2*(N+2)];
  WK[3] = &WK[0][3*(N+2)]; WK[4] = &WK[0][4*(N+2)];
  WK[5] = &WK[0][5*(N+2)]; WK[6] = &WK[0][6*(N+2)];
  SpInt(X,&AVH,F,DF,&AVDF,N,Y,C,&WK[0],&WK[3]); AVAR = VAR;
  if (VAR > 0.0) AVAR = VAR*AVDF*AVDF;
  if (VAR == 0.0) {
    R1 = 0.0;
  } else {
    R1 = 1.0; R2 = Ratio*R1; More = 1;
    SpFit(X,AVH,DF,N,R2,&P,&Q,&GF2,AVAR,STAT,Y,C,&WK[0],&WK[3],WK[5],WK[6]);
    do {
      SpFit(X,AVH,DF,N,R1,&P,&Q,&GF1,AVAR,STAT,Y,C,&WK[0],&WK[3],WK[5],WK[6]);
      if (GF1 > GF2) {
        More = 0;
      } else {
        if (P <= 0.0) {
          More = 0; GoTo100 = 1;
        } else {
          R2 = R1; GF2 = GF1; R1 /= Ratio;
        }
      }
    } while (More);
    if (!GoTo100) {
      R3 = Ratio*R2; More = 1;
      do {
        SpFit(X,AVH,DF,N,R3,&P,&Q,&GF3,AVAR,STAT,Y,C,&WK[0],&WK[3],WK[5],WK[6]);
        if (GF3 > GF2) {
          More = 0;
        } else {
          if (Q <= 0.0) {
            More = 0; GoTo100 = 1;
          } else {
            R2 = R3; GF2 = GF3; R3 = Ratio*R3;
          }
        }
      } while (More);
      R2 = R3; GF2 = GF3; Delta = (R2-R1)/Tau; R4 = R1 + Delta; R3 = R2 - Delta;
      SpFit(X,AVH,DF,N,R3,&P,&Q,&GF3,AVAR,STAT,Y,C,&WK[0],&WK[3],WK[5],WK[6]);
      SpFit(X,AVH,DF,N,R4,&P,&Q,&GF4,AVAR,STAT,Y,C,&WK[0],&WK[3],WK[5],WK[6]);
      do {
        if (GF3 <= GF4) {
          R2 = R4; GF2 = GF4; R4 = R3; GF4 = GF3; Delta /= Tau; R3 = R2 - Delta;
          SpFit(X,AVH,DF,N,R3,&P,&Q,&GF3,AVAR,STAT,Y,C,&WK[0],&WK[3],WK[5],WK[6]);
        } else {
          R1 = R3; GF1 = GF3; R3 = R4; GF3 = GF4; Delta /= Tau; R4 = R1 + Delta;
          SpFit(X,AVH,DF,N,R4,&P,&Q,&GF4,AVAR,STAT,Y,C,&WK[0],&WK[3],WK[5],WK[6]);
        }
        ERR = (R2-R1)/(R1+R2);
      } while (((ERR*ERR+1.0) > 1.0) && (ERR > 1e-6));
      R1 = (R1+R2)*0.5;
    }
  }
  if (!GoTo100) SpFit(X,AVH,DF,N,R1,&P,&Q,&GF1,AVAR,STAT,Y,C,&WK[0],&WK[3],WK[5],WK[6]);
  SpCof(X,AVH,F,DF,N,P,Q,Y,C,WK[5],WK[6]);
  if (VAR < 0.0) {
    AVAR = STAT[5]; VAR = AVAR/(AVDF*AVDF);
  }
  if (Job == 1) SpErr(X,AVH,DF,N,&WK[0],P,AVAR,SE);
  for (i = 0; i < N; i++) DF[i] *= AVDF;
  for (i = 0; i < 6; i++) WK[0][i] = STAT[i];
  WK[0][5] = STAT[5]/(AVDF*AVDF); WK[0][6] = AVDF*AVDF;
}

void SpInt(double *X,double *AvH,double *Y,double *dY,double *AvdY,int N,
           double *A,double *C[3],double *R[3],double *T[2]) {
  int    i;
  double E, F, G = 0.0, H;

  if (N < 3) mexErrMsgTxt("N is less than 3");
  for (i = 0; i < N-1; i++) {
    H = X[i+1] - X[i]; G += H;
    if (H <= 0.0) mexErrMsgTxt("Input abscissae is not monotonic.");
  }
  *AvH = G/(N-1); G = 0.0;
  for (i = 0; i < N; i++) {
    if (dY[i] <= 0.0) mexErrMsgTxt("Errors are not all positive.");
    G += dY[i]*dY[i];
  }
  *AvdY = sqrt(G/N);
  for (i = 0; i < N; i++) dY[i] /= *AvdY;
  H = (X[1]-X[0])/ *AvH; F = (Y[1]-Y[0])/H;
  for (i = 1; i < N-1; i++) {
    G = H; H = (X[i+1]-X[i])/ *AvH; E = F; F = (Y[i+1]-Y[i])/H; A[i] = F - E;
    T[0][i+1] = 2.0*(G+H)/3.0; T[1][i+1] = H/3.0; R[2][i+1] = dY[i-1]/G;
    R[0][i+1] = dY[i+1]/H; R[1][i+1] = -dY[i]/G - dY[i]/H;
  }
  R[1][N] = R[2][N] = R[2][N+1] = 0.0;
  for (i = 1; i < N-1; i++) {
    C[0][i] = R[0][i+1]*R[0][i+1] + R[1][i+1]*R[1][i+1] + R[2][i+1]*R[2][i+1];
    C[1][i] = R[0][i+1]*R[1][i+2] + R[1][i+1]*R[2][i+2];
    C[2][i] = R[0][i+1]*R[2][i+3];
  }
}

void SpFit(double *X,double AvH,double *dY,int N,double Rho,double *P,double *Q,
           double *Fun,double Var,double *Stat,double *A,double *C[3],double *R[3],
           double *T[2],double *U,double *V) {
  int    i;
  double E, F, G, H, Rho1;

  Rho1 = 1.0 + Rho; *P = Rho/Rho1; *Q = 1.0/Rho1; F = G = H = 0.0;

  if (Rho1 == 1.0) *P = 0.0;
  if (Rho1 == Rho) *Q = 0.0;
  R[0][0] = R[0][1] = 0.0;
  for (i = 2; i < N; i++) {
    R[2][i-2] = G*R[0][i-2]; R[1][i-1] = F*R[0][i-1];
    R[0][i] = 1.0/(*P*C[0][i-1] + *Q*T[0][i]-F*R[1][i-1]-G*R[2][i-2]);
    F = *P*C[1][i-1] + *Q*T[1][i] - H*R[1][i-1]; G = H; H = *P*C[2][i-1];
  }
  U[0] = U[1] = U[N] = U[N+1] = E = H = 0.0;
  for (i = 2; i < N; i++) U[i] = A[i-1] - R[1][i-1]*U[i-1] - R[2][i-2]*U[i-2];
  for (i = N-1; i > 1; i--) U[i] = R[0][i]*U[i] - R[1][i]*U[i+1] - R[2][i]*U[i+2];
  for (i = 1; i < N; i++) {
    G = H; H = (U[i+1]-U[i])/((X[i]-X[i-1])/AvH);
    V[i] = dY[i-1]*(H-G); E += V[i]*V[i];
  }
  V[N] = -dY[N-1]*H; E += V[N]*V[N];
  R[0][N] = R[1][N] = R[0][N+1] = 0.0;
  for (i = N-1; i > 1; i--) {
    G = R[1][i]; H = R[2][i]; R[1][i] = -G*R[0][i+1] - H*R[1][i+1];
    R[2][i] = -G*R[1][i+1] - H*R[0][i+1]; R[0][i] -= G*R[1][i] + H*R[2][i];
  }
  F = G = H = 0.0;
  for (i = 2; i < N; i++) {
    F += R[0][i]*C[0][i-1]; G += R[1][i]*C[1][i-1]; H += R[2][i]*C[2][i-1];
  }
  F += 2.0*(G+H);
  Stat[0] = *P; Stat[1] = *P*F; Stat[2] = N*E/(F*F);
  Stat[3] = E* *P* *P/N; Stat[5] = *P*E/F;
  if (Var < 0.0) {
    Stat[4] = Stat[5] - Stat[3]; *Fun = Stat[3];
  } else {
    Stat[4] = Stat[3]-2.0*Var*Stat[1]/N+Var;
    if (Stat[4] < 0.0) Stat[4] = 0.0;
    *Fun = Stat[4];
  }
}

void SpErr(double *X,double AvH,double *dY,int N,double *R[3],double P,
           double Var,double *SE) {
  int    i;
  double F, G, H, F1, G1, H1;

  H = AvH/(X[1]-X[0]); SE[0] = 1.0 - P*dY[0]*dY[0]*H*H*R[0][2];
  R[0][1] = R[1][1] = R[2][1] = 0.0;
  for (i = 2; i < N; i++) {
    F = H; H = AvH/(X[i]-X[i-1]); G = -F - H;
    F1 = F*R[0][i-1] + G*R[1][i-1] + H*R[2][i-1];
    G1 = F*R[1][i-1] + G*R[0][i] + H*R[1][i];
    H1 = F*R[2][i-1] + G*R[1][i] + H*R[0][i+1];
    SE[i-1] = 1.0 - P*dY[i-1]*dY[i-1]*(F*F1+G*G1+H*H1);
  }
  SE[N-1] = 1.0 - P*dY[N-1]*dY[N-1]*H*H*R[0][N-1];
  for (i = 0; i < N; i++) {
    SE[i] *= Var; if (SE[i] < 0.0) SE[i] = 0; SE[i] = sqrt(SE[i])*dY[i];
  }
}

void SpCof(double *X,double AvH,double *Y,double *dY,int N,double P,double Q,
           double *A,double *C[3],double *U,double *V) {
  int    i;
  double H, QH;

  QH = Q/(AvH*AvH);
  for (i = 0; i < N; i++) {
    A[i] = Y[i] - P*dY[i]*V[i+1]; U[i+1] *= QH;
  }
  for (i = 0; i < N-1; i++) {
    H = X[i+1] - X[i];
    C[2][i] = (U[i+2]-U[i+1])/(3.0*H); C[1][i] = U[i+1];
    C[0][i] = (A[i+1]-A[i])/H - (H*C[2][i]+U[i+1])*H;
  }
}
